package SoftLib.core;

import JaCoP.constraints.*;
import JaCoP.core.IntVar;
import JaCoP.core.Store;
import SoftLib.constraints.*;
import SoftLib.weight.PrimitiveBinaryWeight;
import SoftLib.weight.PrimitiveUnaryWeight;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

/*
 * This class must be initilized to work with soft contraint
 * it extends JaCoP's Store class and add to it:
 * - a "costs" list which contains costs of constrait
 * - a "srType" variable which represent the semiring type
 * - "getCosts()" method which to compute the totatal costs according the semiring type
 */
public class SoftStore extends Store {

    public ArrayList<IntVar> costs;
    public ArrayList<ExBinaryReference> binary;
    public HashMap<IntVar,ArrayList<CoupleConstraint>> unary;
    protected WeightStore wStore;
    private Boolean debugMode = false;

    public enum Semiring {
        WEIGHTED,
        PROBABILISTIC,
        FUZZY
    }
    private Semiring srType;
    
    /*
     * Empty costructor: WEIGHTED semiring type will be used as default
     */
    public SoftStore() {
        this(Semiring.WEIGHTED);
    }
    
    /*
     * Costructo with one parameter: set the given semiring type
     */
    public SoftStore(Semiring srType) {
        super();
        costs = new ArrayList<IntVar>();
        binary = new ArrayList<ExBinaryReference>();
        unary = new HashMap<IntVar, ArrayList<CoupleConstraint>>();
        this.srType = srType;
    }

    public Semiring getSrType() {
        return srType;
    }

    public WeightStore getwStore() {
        return wStore;
    }
    
    public void setDebugMode (Boolean active) {
        this.debugMode = active;
    }

    /*
     * impose() method for generic crisp constraint
     */
    @Override
    public void impose(Constraint c) {
        c.impose(this);
    }

    /*
     * impose() method for SoftXneqY constraint with PrimitiveWeight object
     * parameter. In this class will be implemented the cost function which can
     * be specified by programmer or can be one of those already defined in the
     * examples
     */
    public void impose(SoftXneqY c, PrimitiveBinaryWeight w) {
        c.impose(this, w);
    }
    
    public void impose(SoftXltY c, PrimitiveBinaryWeight w) {
        c.impose(this, w);
    }
    
    public void impose(SoftXeqC c, PrimitiveUnaryWeight w) {
        c.impose(this, w);
    }
    
    public void impose(SoftXneqC c, PrimitiveUnaryWeight w) {
        c.impose(this, w);
    }
    
    public void impose(SoftXltC c, PrimitiveUnaryWeight w) {
        c.impose(this, w);
    }

    /*
     * impose() method for SoftExtensionalSupport method with PrimitiveWeight 
     * object parameter. In this class will be implemented the cost function 
     * which can be specified by programmer or can be one of those already 
     * defined in the examples.
     * (this method can't be interfaced)
     */
    public void impose(SoftExtensionalSupportVA c, PrimitiveBinaryWeight w) {
        c.impose(this, w);
    }
    
    public void impose(SoftExtensionalSupportSTR c, PrimitiveBinaryWeight w) {
        c.impose(this, w);
    }
    
    public void impose(SoftExtensionalSupportMDD c, PrimitiveBinaryWeight w) {
        c.impose(this, w);
    }

    /*
     * getCost() method: it computes and returns costs
     */
    public IntVar getCost() {
        
        /* 
         * the totalCost variable stores the total costs computed according 
         * the semiring type
         */
        IntVar totalCost, result;
        /*
         * totCostNegative variable stores the totalCost[0] moltiplied by -1
         * to maximize it
         */
        IntVar totCostNegative;
        /*
         * costMult is used to compute moltiplication in PROBABILISTC semirings:
         * in JaCoP does not exists a constraint able to compute more 
         * moltiplications at a time, so in this array are stored partial results
         */
        IntVar[] costMult;
        switch (srType) {
            case FUZZY:
                /*
                 * the total cost is calculated as the minimum between all costs. the costs is
                 * then given to the labeling which will minimize it
                 */
                int min = costs.get(0).min();
                int max = costs.get(0).max();
                for (int i = 1; i < costs.size(); i++) {
                    if (costs.get(i).min() < min) {
                        min = costs.get(i).min();
                    }
                    if (costs.get(i).max() < max) {
                        max = costs.get(i).max();
                    }
                }
                totalCost = new IntVar(this, "totalCost", min, max);
                this.impose(new Min(costs, totalCost));
                totCostNegative = new IntVar(this, "totalCostNegative", -1*totalCost.max(), -1*totalCost.min());
                this.impose(new XmulCeqZ(totalCost, -1, totCostNegative));
                result = totCostNegative;
                break;
            case PROBABILISTIC:
                /*
                 * the total cost is calculated as the product between costs. First of all
                 * costMult[] array is created, then will be set moltiplication
                 * constraints and returned to the labeling
                 */
                IntVar factor =  new IntVar(this, "factor", 1000, 1000);
                int min2 = costs.get(0).min();
                int max2 = costs.get(0).max();
                for (int i = 1; i < costs.size(); i++) {
                    min2 = (min2*costs.get(i).min())/1000;
                    max2 = (max2*costs.get(i).max())/1000;
                }
                totalCost = new IntVar(this, "totalCost", min2, max2);
                if (costs.size() > 1) {
                    costMult = new IntVar[(costs.size() - 1)*2];
                    costMult[0] = new IntVar(this, "mult", costs.get(0).min()*costs.get(1).min(), costs.get(0).max()*costs.get(1).max());
                } else{
                    costMult = new IntVar[1];
                    costMult[0] = new IntVar(this, "mult", costs.get(0).min(), costs.get(0).max());
                }    
                costMult[1] = new IntVar(this, "mult", costMult[0].min()/1000, costMult[0].max()/1000);
                if (costs.size() > 1) {
                    this.impose(new XmulYeqZ(costs.get(0), costs.get(1), costMult[0]));
                } else if (costs.size() == 1) {
                    this.impose(new XeqY(costs.get(0), costMult[0]));
                }
                this.impose(new XdivYeqZ(costMult[0],factor,costMult[1]));
                for (int i = 2; i < costs.size(); i++) {
                    costMult[2*i-2] = new IntVar(this, "mult", costs.get(i).min()*costMult[2*i-3].min(), costs.get(i).max()*costMult[2*i-3].max());
                    this.impose(new XmulYeqZ(costMult[2*i-3], costs.get(i), costMult[2*i-2]));
                    costMult[2*i-1] = new IntVar(this, "mult", costMult[2*i-2].min()/1000, costMult[2*i-2].max()/1000);
                    this.impose(new XdivYeqZ(costMult[2*i-2],factor,costMult[2*i-1]));
                }
                this.impose(new XeqY(costMult[costMult.length - 1], totalCost));
                totCostNegative = new IntVar(this, "totalCostNegative", -1*totalCost.max(), -1*totalCost.min());
                this.impose(new XmulCeqZ(totalCost, -1, totCostNegative));
                result = totCostNegative;
                break;
            case WEIGHTED:
                /*
                 * the total cost is calculated as the sum of all costs, the value is then 
                 * returned to the labeling which will minimize it
                 */
                int min3 = costs.get(0).min();
                int max3 = costs.get(0).max();
                for (int i = 1; i < costs.size(); i++) {
                    min3 += costs.get(i).min();
                    max3 += costs.get(i).max();
                }
                totalCost = new IntVar(this, "totalCost", min3, max3);
                this.impose(new Sum(costs, totalCost));
                result = totalCost;
                break;
            default:
                throw new UnsupportedOperationException("Not supported type: choose a valid semiring type. FUZZY, PROBABILISTIC or WEIGHTED.");
        }
        return result;
    }
    
    /**
     * Function used to properly fill the field this.binary.
     */
    public void addToBinary (IntVar x, IntVar y, CoupleConstraint c){
        Iterator<ExBinaryReference> iter = binary.iterator();
        ExBinaryReference currentBinary;
        boolean added = false;
        while(iter.hasNext() && added == false){ //iterates on binary...
            currentBinary = iter.next();
            if(currentBinary.getNodex() == x && currentBinary.getNodey() == y) { // ...searching for an ExBinaryReference for x and y
                currentBinary.constraints.add(c);
                added = true;
            }
        }
        if(!added){ //if there is not yet an ExBinaryReference related to x and y, we musst create it
            ExBinaryReference newExBinary = new ExBinaryReference(x, y);
            newExBinary.constraints.add(c);
            binary.add(newExBinary);
        }
    }
    
    /**
     * Function used to properly fill the field this.unary.
     */
    public void addToUnary (IntVar x, CoupleConstraint c){        
        if(unary.containsKey(x)){
            unary.get(x).add(c);
        } else {
            ArrayList<CoupleConstraint> list = new ArrayList<CoupleConstraint>();
            list.add(c);
            unary.put(x, list);
        }
    }
    
    /**
     * Function used to translate weight constraints in a more convenient numeric form.
     * It takes every unary constraint contained in this.unary and, for each of them,
     * it stores in wstore a new hashMap.
     * The hashMap contains the domain value of the variable as key and the related weight as value.
     */
    private void formatUnaryWeight(WeightStore wstore){

        this.costs.clear();
        Iterator<IntVar> itIntVar = unary.keySet().iterator();
        IntVar currentIntVar;
        WeightUnaryDomain wUnary;
        
        Iterator<CoupleConstraint> itCon;
        CoupleConstraint currentCC;
        
        Iterator<Integer> itDom;
        HashMap<Integer,Integer> currentHM;
        int currentDom;
        int currentWeight;
        
        while(itIntVar.hasNext()){
            currentIntVar = itIntVar.next();
            wUnary = new WeightUnaryDomain(currentIntVar);
            wstore.wUnary.put(currentIntVar, wUnary);
            wUnary.w = new ArrayList<HashMap<Integer,Integer>>();
            itCon = unary.get(currentIntVar).iterator();
            while(itCon.hasNext()){
                currentCC = itCon.next();
                currentHM = new HashMap<Integer, Integer>();
                itDom = wUnary.xDom.iterator();
                while(itDom.hasNext()){
                    currentDom = itDom.next();
                    currentWeight = ((PrimitiveUnaryWeight)currentCC.getConstraint().getWeight()).evaluateWeight(currentDom);
                    currentHM.put(currentDom, currentWeight);
                }
                wUnary.w.add(currentHM);
            }
        }
    }
    
    /**
     * Function used to translate weight constraints in a more convenient numeric form.
     * It takes every binary constraint contained in this.binary and, for each couple of variables
     * linked by at least one constraint it creates a list of HashMap.
     * Every HashMap in the list represent a constraint in the form domainValue -> weightValue.
     */
    private void formatBinaryWeight(WeightStore wstore){
        
        Iterator<ExBinaryReference> itBinary = binary.iterator();
        ExBinaryReference currentExBinary;
        WeightBinaryDomains xyDom;
        WeightUnaryDomain xDom, yDom;

        Iterator<CoupleConstraint> itCon;
        CoupleConstraint currentCoupleConstraint;
        HashMap<IntCouple,Integer> currentHM;
        
        Iterator<Integer> itX, itY;
        int localX, localY, w;
        
        while(itBinary.hasNext()){ //iterate on the lists of couple of variables
            currentExBinary = itBinary.next();
            xyDom = new WeightBinaryDomains(currentExBinary.getNodex(), currentExBinary.getNodey());
            wstore.wBinary.add(xyDom);
            xDom = wstore.wUnary.get(xyDom.x);
            yDom = wstore.wUnary.get(xyDom.y);
            itCon = currentExBinary.constraints.iterator();
            while(itCon.hasNext()){ //iterate on the list of soft constraints related to the current couple
                currentCoupleConstraint = itCon.next();
                currentHM = new HashMap<IntCouple, Integer>();
                /*
                 *  now we have to generate every combinations
                 *  of couple's domains elements and calculate
                 *  the weight using the 'evalutateWeight'
                 *  method of this constraint
                 */
                itX = xDom.xDom.iterator();
                while(itX.hasNext()){
                    localX = itX.next();
                    itY = yDom.xDom.iterator();
                    while(itY.hasNext()){    
                        localY = itY.next();
                        w = ((PrimitiveBinaryWeight)currentCoupleConstraint.getConstraint().getWeight()).evaluateWeight(localX, localY);
                        currentHM.put(new IntCouple(localX, localY), w);
                    }
                }
                xyDom.w.add(currentHM);
            }
        }
    }
    
    /**
     * Function used to merge unary weight constraints related to the same variables.
     */
    public void weightUnaryFilter (WeightStore wStore) {
        Set<IntVar> keys = wStore.wUnary.keySet();
        Iterator<IntVar> itVar = keys.iterator();
        Iterator<HashMap<Integer, Integer>> itHash;
        Iterator<Integer> itInteger;
        IntVar currentIntVar;
        HashMap<Integer, Integer> currentHash;
        WeightUnaryDomain currentUnary;
        int domValue, weightValue, hashID;
        int domMin, domMax;
        IntVar weightVar;
        IntVar[] varsUnary = new IntVar[2];
        int[][] supported;
        while (itVar.hasNext()) {
            currentIntVar = itVar.next();
            currentUnary = wStore.wUnary.get(currentIntVar);
            itHash = currentUnary.w.iterator();
            if (itHash.hasNext()) {
                currentHash = itHash.next();
                supported = new int[currentHash.size()][2];
                hashID = 0;
                itInteger = currentUnary.xDom.iterator();
                while (itInteger.hasNext()) {
                    domValue = itInteger.next();
                    weightValue = currentHash.get(domValue);
                    if (weightValue != 0) {
                        supported[hashID][0] = domValue;
                        supported[hashID][1] = weightValue;
                        hashID++;
                    }
                }
                int[][] subsupported = new int[hashID][2];
                System.arraycopy(supported, 0, subsupported, 0, hashID);
                supported = subsupported;
                while (itHash.hasNext()) {
                    currentHash = itHash.next();
                    hashID = 0;
                    itInteger = currentUnary.xDom.iterator();
                    while (itInteger.hasNext()) {
                        domValue = itInteger.next();
                        weightValue = currentHash.get(domValue);
                        if (weightValue != 0) {
                            if (supported[hashID][1] > weightValue) {
                                supported[hashID][1] = weightValue; 
                            }
                            hashID++;
                        }
                    }
                }
                domMin = Integer.MAX_VALUE;
                domMax = Integer.MIN_VALUE;
                for(int i=0; i < supported.length; i++){
                    if(domMax < supported[i][1]){ domMax = supported[i][1];}
                    if(domMin > supported[i][1]){ domMin = supported[i][1];}
                }
                if (this.debugMode) {
                    System.out.println("UNARY domMin: "+domMin+" && domMax: "+domMax);
                }
                weightVar = new IntVar(this, domMin, domMax);
                varsUnary[0] = currentIntVar;
                varsUnary[1] = weightVar;
                costs.add(varsUnary[1]);
                currentUnary.weightCon = new ExtensionalSupportSTR(varsUnary, supported);
            }
        }
    }

    /**
     * Function used to merge binary constraints related to the same couple of variables.
     */
    public void weightBinaryFilter (WeightStore wStore) {
        Iterator<WeightBinaryDomains> itWBD = wStore.wBinary.iterator();
        WeightBinaryDomains currentWBD;
        Iterator<HashMap<IntCouple, Integer>> itHash;
        HashMap<IntCouple, Integer> currentHash;
        Set<IntCouple> keys;
        Iterator<IntCouple> itIntCouple;
        IntCouple currentIntCouple;
        IntVar[] varsBinary = new IntVar[3];
        int[][]supported;
        int hashID;
        int domMin, domMax;
        while(itWBD.hasNext()){
            currentWBD = itWBD.next();
            itHash = currentWBD.w.iterator();
            if(itHash.hasNext()){
                currentHash = itHash.next();
                keys = currentHash.keySet();
                supported = new int[keys.size()][3];
                itIntCouple = keys.iterator();
                hashID = 0;
                while(itIntCouple.hasNext()){
                    currentIntCouple = itIntCouple.next();
                    supported[hashID][0] = currentIntCouple.x;
                    supported[hashID][1] = currentIntCouple.y;
                    supported[hashID][2] = currentHash.get(currentIntCouple);
                    hashID++;
                }
                while(itHash.hasNext()){
                    currentHash = itHash.next();
                    itIntCouple = keys.iterator();
                    hashID = 0;
                    while(itIntCouple.hasNext()){
                        currentIntCouple = itIntCouple.next();
                        if(currentHash.get(currentIntCouple) < supported[hashID][2]){
                            supported[hashID][2] = currentHash.get(currentIntCouple);
                        }
                        hashID++;
                    }
                }
                domMin = Integer.MAX_VALUE;
                domMax = Integer.MIN_VALUE;
                for(int i=0; i < supported.length; i++){
                    if(domMax < supported[i][2]){ domMax = supported[i][2];}
                    if(domMin > supported[i][2]){ domMin = supported[i][2];}
                }
                if (this.debugMode) {
                    System.out.println("BINARY domMin: "+domMin+" && domMax: "+domMax);
                }
                varsBinary[0] = currentWBD.x;
                varsBinary[1] = currentWBD.y;
                varsBinary[2] = new IntVar(this, domMin, domMax);
                costs.add(varsBinary[2]);
                currentWBD.weightCon = new ExtensionalSupportSTR(varsBinary, supported);
            }
        }
    }
            
    /**
     * AC-1 based soft arc consistency procedures.
     * In order to propagate to the search algorithm the results computed in this function,
     * in the last lines, every domain value with a weight value equals to 0, is removed.
     */
    public void softArcConsistency() {
        if (this.srType != Semiring.FUZZY) {
            throw new UnsupportedOperationException("Wrong semiring type");
        }
        wStore = new WeightStore();
        formatUnaryWeight(wStore);
        if (this.debugMode) {
            System.out.println("formatUnaryWeight() DONE");
        }
        formatBinaryWeight(wStore);
        if (this.debugMode) {
            System.out.println("formatBinaryWeight() DONE");
        }
        WeightBinaryDomains xyDomain;
        WeightUnaryDomain xDomain, yDomain;
        Iterator<WeightBinaryDomains> itWB;
        Iterator<HashMap<Integer, Integer>> itWeightX, itWeightY;
        Iterator<HashMap<IntCouple, Integer>> itWeightXY;
        HashMap<Integer, Integer> hashX = null, hashY = null;
        HashMap<IntCouple, Integer> hashXY = null;
        int weightX, weightY, weightXY, weightMin, weightMax;
        IntCouple supportKey = new IntCouple(0, 0);
        boolean localChanged = true, done;
        while(localChanged){
            localChanged = false;
            itWB = wStore.wBinary.iterator();
            while (itWB.hasNext()) { //iterates on the couple of variables involved in at least one binary constraint.
                xyDomain = itWB.next();
                xDomain = wStore.wUnary.get(xyDomain.x);
                yDomain = wStore.wUnary.get(xyDomain.y);
                itWeightXY = xyDomain.w.iterator();
                itWeightX = xDomain.w.iterator();
                itWeightY = yDomain.w.iterator();
                done = false;
                while (!done) {
                    done = true;
                    if (itWeightXY.hasNext()) { //select a new binary constraint if there is
                        hashXY = itWeightXY.next();
                        done = false;
                    }
                    if (itWeightX.hasNext()) { //select a new unary constraint for X if there is
                        hashX = itWeightX.next();
                        done = false;
                    }
                    if (itWeightY.hasNext()) { //select a new unary constraint for Y if there is
                        hashY = itWeightY.next();
                        done = false;
                    }
                    if (!done) {
                        for (int indexX = 0; indexX < xDomain.xDom.size(); indexX++) { //for every value x in the domain of X
                            weightX = hashX.get(xDomain.xDom.get(indexX));
                            weightMax = Integer.MIN_VALUE;
                            for (int indexY = 0; indexY < yDomain.xDom.size(); indexY++) { //compute the maximum weight value reachable considering X=x
                                supportKey.setX(xDomain.xDom.get(indexX));
                                supportKey.setY(yDomain.xDom.get(indexY));
                                weightXY = hashXY.get(supportKey);
                                weightY = hashY.get(yDomain.xDom.get(indexY));
                                weightMin = Math.min(Math.min(weightX, weightY), weightXY); //it's the resulting weight from the current varible assignemnt
                                weightMax = Math.max(weightMax, weightMin); //it's the adjusted weight value for the current X variable assignemnt
                            }
                            if (weightX != weightMax) {
                                hashX.put(xDomain.xDom.get(indexX), weightMax);
                                localChanged = true;
                            }
                        }
                        for (int indexY = 0; indexY < yDomain.xDom.size(); indexY++) {//for every value y in the domain of Y
                            weightY = hashY.get(yDomain.xDom.get(indexY));
                            weightMax = Integer.MIN_VALUE;
                            for (int indexX = 0; indexX < xDomain.xDom.size(); indexX++) {//compute the maximum weight value reachable considering Y=y
                                supportKey.setX(xDomain.xDom.get(indexX));
                                supportKey.setY(yDomain.xDom.get(indexY));
                                weightXY = hashXY.get(supportKey);
                                weightX = hashX.get(xDomain.xDom.get(indexX));
                                weightMin = Math.min(Math.min(weightX, weightY), weightXY);
                                weightMax = Math.max(weightMax, weightMin);
                            }
                            if (weightY != weightMax) {
                                hashY.put(yDomain.xDom.get(indexY), weightMax);
                                localChanged = true;
                            }
                        }
                    }
                }
            }
        }     
        if (this.debugMode) {
            System.out.println("soft arc consistency alg DONE");
        }
        weightUnaryFilter(wStore);
        if (this.debugMode) {
            System.out.println("weightUnaryFilter(wStore) DONE");
        }
        weightBinaryFilter(wStore);
        if (this.debugMode) {
            System.out.println("weightBinaryFilter(wStore) DONE");
        }
        binary = null;
        unary = null;
    }
    
    /**
     * Function used to impose weight constraints whether softArcConsistency was called or not.
     */
    public void imposeWeights() {
        Set<IntVar> keys;
        Iterator<IntVar> itIntVar;
        IntVar currentIntVar;
        if(unary != null || binary != null){ //then softArcConsistency() was not called so take the weight constraints from binary and unary and impose them.
            keys = unary.keySet();
            itIntVar = keys.iterator();
            Iterator<CoupleConstraint> itCC;
            CoupleConstraint currentCC;
            while(itIntVar.hasNext()){
                itCC = unary.get(itIntVar.next()).iterator();
                while(itCC.hasNext()){
                    currentCC = itCC.next();
                    currentCC.getCost().impose(this);
                }
            }
            Iterator<ExBinaryReference> itEBR;
            ExBinaryReference currentEBR;
            itEBR = binary.iterator();
            while(itEBR.hasNext()){
                currentEBR = itEBR.next();
                itCC = currentEBR.constraints.iterator();
                while(itCC.hasNext()){
                    currentCC = itCC.next();
                    currentCC.getCost().impose(this);
                }
            }
        } else {//softArcConsistency was called, so weight constraint to impose are stored in wStore.
            keys = wStore.wUnary.keySet();
            itIntVar = keys.iterator();
            WeightUnaryDomain currentWUD;
            while(itIntVar.hasNext()){
                currentIntVar = itIntVar.next();
                currentWUD = wStore.wUnary.get(currentIntVar);
                currentWUD.weightCon.impose(this);
            }

            Iterator<WeightBinaryDomains> itWBD = wStore.wBinary.iterator();
            WeightBinaryDomains currentWBD;
            while(itWBD.hasNext()){
                currentWBD = itWBD.next();
                currentWBD.weightCon.impose(this);
            }
            wStore = null;
        }
    }
    
}